"""Tests for datset."""

import unittest
import os

import numpy as np

import aspecd.analysis
import aspecd.annotation
import aspecd.exceptions
import aspecd.history
import aspecd.plotting
import aspecd.processing
import aspecd.utils
from aspecd import (
    annotation,
    analysis,
    dataset,
    io,
    plotting,
    processing,
    table,
)
import aspecd.metadata


class TestDataset(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()

    def test_instantiate_class(self):
        pass

    def test_has_data_property(self):
        self.assertTrue(hasattr(self.dataset, "data"))

    def test_data_is_data(self):
        self.assertTrue(isinstance(self.dataset.data, dataset.Data))

    def test_has_origdata_property(self):
        self.assertTrue(hasattr(self.dataset, "_origdata"))

    def test_origdata_is_data(self):
        self.assertTrue(isinstance(self.dataset._origdata, dataset.Data))

    def test_has_metadata_property(self):
        self.assertTrue(hasattr(self.dataset, "metadata"))

    def test_has_history_property(self):
        self.assertTrue(hasattr(self.dataset, "history"))

    def test_history_is_list(self):
        self.assertTrue(isinstance(self.dataset.history, list))

    def test_has_analyses_property(self):
        self.assertTrue(hasattr(self.dataset, "analyses"))

    def test_analyses_is_list(self):
        self.assertTrue(isinstance(self.dataset.analyses, list))

    def test_has_protected_package_name_property(self):
        self.assertTrue(hasattr(self.dataset, "_package_name"))

    def test_has_representations_property(self):
        self.assertTrue(hasattr(self.dataset, "representations"))

    def test_has_id_property(self):
        self.assertTrue(hasattr(self.dataset, "id"))

    def test_has_label_property(self):
        self.assertTrue(hasattr(self.dataset, "label"))

    def test_has_references_property(self):
        self.assertTrue(hasattr(self.dataset, "references"))

    def test_has_package_name_property(self):
        self.assertTrue(hasattr(self.dataset, "package_name"))

    def test_package_name_property_is_readonly(self):
        with self.assertRaises(AttributeError):
            # noinspection PyPropertyAccess
            self.dataset.package_name = "foo"

    def test_has_tasks_property(self):
        self.assertTrue(hasattr(self.dataset, "tasks"))

    def test_has_device_data_property(self):
        self.assertTrue(hasattr(self.dataset, "device_data"))

    def test_device_data_is_dict(self):
        self.assertTrue(isinstance(self.dataset.device_data, dict))


class TestDatasetProcessing(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.processing_step = processing.SingleProcessingStep()

    def test_has_process_method(self):
        self.assertTrue(hasattr(self.dataset, "process"))
        self.assertTrue(callable(self.dataset.process))

    def test_process_without_processing_step_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingProcessingStepError):
            self.dataset.process()

    def test_process_adds_history_record(self):
        self.dataset.process(self.processing_step)
        self.assertFalse(self.dataset.history == [])

    def test_added_history_record_is_historyrecord(self):
        self.dataset.process(self.processing_step)
        self.assertTrue(
            isinstance(
                self.dataset.history[-1],
                aspecd.history.ProcessingHistoryRecord,
            )
        )

    def test_process_increments_history_pointer(self):
        historypointer = self.dataset._history_pointer
        self.dataset.process(self.processing_step)
        self.assertTrue(self.dataset._history_pointer == historypointer + 1)

    def test_process_writes_different_historyrecords(self):
        self.dataset.process(self.processing_step)
        self.dataset.process(self.processing_step)
        self.assertIsNot(
            self.dataset.history[-1].processing,
            self.dataset.history[-2].processing,
        )

    def test_process_history_record_process_is_processing_step_record(self):
        self.dataset.process(self.processing_step)
        self.assertTrue(
            isinstance(
                self.dataset.history[-1].processing,
                aspecd.history.ProcessingStepRecord,
            )
        )

    def test_process_copies_processingstep_object(self):
        self.dataset.process(self.processing_step)
        self.assertIsNot(
            self.processing_step, self.dataset.history[-1].processing
        )

    def test_process_returns_processing_object(self):
        processing_object = processing.SingleProcessingStep()
        processing_step = self.dataset.process(processing_object)
        self.assertTrue(
            isinstance(processing_step, processing.SingleProcessingStep)
        )

    def test_process_sets_package_in_sysinfo(self):
        # Fake package name
        self.dataset._package_name = "numpy"
        self.dataset.process(self.processing_step)
        history_record = self.dataset.history[0]
        self.assertTrue("numpy" in history_record.sysinfo.packages.keys())

    def test_undoable_processing_step_does_not_touch_origdata(self):
        processing_step = processing.SingleProcessingStep()
        processing_step.undoable = True
        old_origdata = self.dataset._origdata
        self.dataset.process(processing_step)
        self.assertIs(self.dataset._origdata, old_origdata)

    def test_not_undoable_processing_step_resets_origdata(self):
        processing_step = processing.SingleProcessingStep()
        processing_step.undoable = False
        old_origdata = self.dataset._origdata
        self.dataset.process(processing_step)
        self.assertIsNot(self.dataset._origdata, old_origdata)

    def test_not_undoable_processing_step_empties_representations(self):
        processing_step = processing.SingleProcessingStep()
        processing_step.undoable = False
        self.dataset.process(processing_step)
        self.assertEqual(self.dataset.representations, [])

    def test_process_adds_task(self):
        self.dataset.process(self.processing_step)
        self.assertNotEqual(self.dataset.tasks, [])

    def test_added_task_has_kind_processing(self):
        self.dataset.process(self.processing_step)
        self.assertEqual(self.dataset.tasks[0]["kind"], "processing")

    def test_added_task_has_processing_history_record(self):
        self.dataset.process(self.processing_step)
        self.assertIsInstance(
            self.dataset.tasks[0]["task"],
            aspecd.history.ProcessingHistoryRecord,
        )


class TestDatasetUndo(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.processingStep = processing.SingleProcessingStep()

    def test_has_undo_method(self):
        self.assertTrue(hasattr(self.dataset, "undo"))
        self.assertTrue(callable(self.dataset.undo))

    def test_undo_with_empty_history_raises(self):
        self.dataset.history.clear()
        with self.assertRaises(aspecd.exceptions.UndoWithEmptyHistoryError):
            self.dataset.undo()

    def test_undo_decrements_historypointer(self):
        self.dataset.process(self.processingStep)
        historypointer = self.dataset._history_pointer
        self.dataset.undo()
        self.assertEqual(self.dataset._history_pointer, historypointer - 1)

    def test_undo_with_historypointer_zero_raises(self):
        self.dataset.process(self.processingStep)
        self.dataset.undo()
        with self.assertRaises(
            aspecd.exceptions.UndoAtBeginningOfHistoryError
        ):
            self.dataset.undo()

    def test_undo_with_undoable_processing_step_raises(self):
        processingstep = self.processingStep
        processingstep.undoable = True
        self.dataset.process(processingstep)
        with self.assertRaises(aspecd.exceptions.UndoStepUndoableError):
            self.dataset.undo()

    def test_multiple_undo_with_undoable_processing_step_raises(self):
        processingstep = self.processingStep
        processingstep.undoable = True
        self.dataset.process(processingstep)
        processingstep.undoable = False
        self.dataset.process(processingstep)
        self.dataset.undo()
        with self.assertRaises(aspecd.exceptions.UndoStepUndoableError):
            self.dataset.undo()


class TestDatasetRedo(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.processingStep = processing.SingleProcessingStep()

    def test_has_redo_method(self):
        self.assertTrue(hasattr(self.dataset, "redo"))
        self.assertTrue(callable(self.dataset.redo))

    def test_redo_with_empty_history_raises(self):
        self.dataset.history.clear()
        with self.assertRaises(
            aspecd.exceptions.RedoAlreadyAtLatestChangeError
        ):
            self.dataset.redo()

    def test_redo_at_latest_change_raises(self):
        self.dataset.process(self.processingStep)
        with self.assertRaises(
            aspecd.exceptions.RedoAlreadyAtLatestChangeError
        ):
            self.dataset.redo()

    def test_redo_increments_historypointer(self):
        self.dataset.process(self.processingStep)
        self.dataset.undo()
        history_pointer = self.dataset._history_pointer
        self.dataset.redo()
        self.assertEqual(self.dataset._history_pointer, history_pointer + 1)


class TestDatasetIO(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.filename = "foo"

    def tearDown(self):
        if os.path.exists(self.filename + ".adf"):
            os.remove(self.filename + ".adf")

    def test_has_load_method(self):
        self.assertTrue(hasattr(self.dataset, "load"))
        self.assertTrue(callable(self.dataset.load))

    def test_has_save_method(self):
        self.assertTrue(hasattr(self.dataset, "save"))
        self.assertTrue(callable(self.dataset.save))

    def test_has_import_from_method(self):
        self.assertTrue(hasattr(self.dataset, "import_from"))
        self.assertTrue(callable(self.dataset.import_from))

    def test_has_export_to_method(self):
        self.assertTrue(hasattr(self.dataset, "export_to"))
        self.assertTrue(callable(self.dataset.export_to))

    def test_import_from_sets_origdata(self):
        importer = io.DatasetImporter()
        old_origdata = self.dataset._origdata
        self.dataset.import_from(importer)
        self.assertIsNot(self.dataset._origdata, old_origdata)

    def test_save_writes_file(self):
        self.dataset.save(filename=self.filename)
        self.assertTrue(os.path.exists(self.filename + ".adf"))

    def test_load_sets_data(self):
        self.dataset.data.data = np.random.random(10)
        self.dataset.save(filename=self.filename)
        new_dataset = aspecd.dataset.Dataset()
        new_dataset.load(filename=self.filename)
        np.testing.assert_allclose(
            self.dataset.data.data, new_dataset.data.data
        )


class TestDatasetProcessingWithHistory(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.processingStep = processing.SingleProcessingStep()
        self.dataset.process(self.processingStep)
        self.dataset.undo()

    def test_stripping_leading_history_deletes_history_entries(self):
        orig_len_history = len(self.dataset.history)
        self.dataset.strip_history()
        new_len_history = len(self.dataset.history)
        self.assertGreater(orig_len_history, new_len_history)

    def test_process_with_leading_history_raises(self):
        with self.assertRaises(
            aspecd.exceptions.ProcessingWithLeadingHistoryError
        ):
            self.dataset.process(self.processingStep)

    def test_stripping_leading_history_allows_processing(self):
        self.dataset.strip_history()
        self.dataset.process(self.processingStep)


class TestDatasetAnalysis(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.analysis_step = analysis.SingleAnalysisStep()

    def test_has_analyse_method(self):
        self.assertTrue(hasattr(self.dataset, "analyse"))
        self.assertTrue(callable(self.dataset.analyse))

    def test_has_analyze_method(self):
        self.assertTrue(hasattr(self.dataset, "analyze"))
        self.assertTrue(callable(self.dataset.analyze))

    def test_analyse_adds_analysis_record(self):
        self.dataset.analyse(self.analysis_step)
        self.assertFalse(self.dataset.analyses == [])

    def test_added_analysis_record_is_analysishistoryrecord(self):
        self.dataset.analyse(self.analysis_step)
        self.assertTrue(
            isinstance(
                self.dataset.analyses[-1],
                aspecd.history.AnalysisHistoryRecord,
            )
        )

    def test_added_analysis_record_contains_history(self):
        processing_step = processing.SingleProcessingStep()
        self.dataset.process(processing_step)
        self.dataset.analyse(self.analysis_step)
        analysis_ = self.dataset.analyses[-1]
        self.assertEqual(
            len(self.dataset.history), len(analysis_.analysis.preprocessing)
        )

    def test_added_analysis_record_history_is_deepcopy(self):
        processing_step = processing.SingleProcessingStep()
        self.dataset.process(processing_step)
        self.dataset.analyse(self.analysis_step)
        analysis_ = self.dataset.analyses[-1]
        self.assertNotEqual(
            self.dataset.history, analysis_.analysis.preprocessing
        )

    def test_analyse_returns_analysis_object(self):
        analysis_step = self.dataset.analyse(self.analysis_step)
        self.assertTrue(
            isinstance(analysis_step, analysis.SingleAnalysisStep)
        )

    def test_analyze_returns_analysis_object(self):
        analysis_step = self.dataset.analyze(self.analysis_step)
        self.assertTrue(
            isinstance(analysis_step, analysis.SingleAnalysisStep)
        )

    def test_has_delete_analysis_method(self):
        self.assertTrue(hasattr(self.dataset, "delete_analysis"))
        self.assertTrue(callable(self.dataset.delete_analysis))

    def test_delete_analysis_deletes_analysis_record(self):
        self.dataset.analyse(self.analysis_step)
        orig_len_analyses = len(self.dataset.analyses)
        self.dataset.delete_analysis(0)
        new_len_analyses = len(self.dataset.analyses)
        self.assertGreater(orig_len_analyses, new_len_analyses)

    def test_delete_analysis_deletes_correct_analysis_record(self):
        self.dataset.analyse(self.analysis_step)
        self.dataset.analyse(self.analysis_step)
        analysisstep = self.dataset.analyses[-1]
        self.dataset.delete_analysis(0)
        self.assertIs(analysisstep, self.dataset.analyses[-1])

    def test_dataset_analyse_sets_package_in_sysinfo(self):
        # Fake package name
        self.dataset._package_name = "numpy"
        self.dataset.analyse(self.analysis_step)
        analysis_record = self.dataset.analyses[0]
        self.assertTrue("numpy" in analysis_record.sysinfo.packages.keys())

    def test_analyse_adds_task(self):
        self.dataset.analyse(self.analysis_step)
        self.assertNotEqual(self.dataset.tasks, [])

    def test_added_task_has_kind_analysis(self):
        self.dataset.analyse(self.analysis_step)
        self.assertEqual(self.dataset.tasks[0]["kind"], "analysis")

    def test_added_task_has_analysis_history_record(self):
        self.dataset.analyse(self.analysis_step)
        self.assertIsInstance(
            self.dataset.tasks[0]["task"],
            aspecd.history.AnalysisHistoryRecord,
        )


class TestDatasetAnnotation(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.annotation = annotation.DatasetAnnotation()
        self.annotation.content = "boo"

    def test_has_annotate_method(self):
        self.assertTrue(hasattr(self.dataset, "annotate"))
        self.assertTrue(callable(self.dataset.annotate))

    def test_annotate_adds_annotation_record(self):
        self.dataset.annotate(self.annotation)
        self.assertFalse(self.dataset.annotations == [])

    def test_added_annotation_record_is_datasetannotationhistoryrecord(self):
        self.dataset.annotate(self.annotation)
        self.assertIsInstance(
            self.dataset.annotations[-1],
            aspecd.history.AnnotationHistoryRecord,
        )

    def test_has_delete_annotation_method(self):
        self.assertTrue(hasattr(self.dataset, "delete_annotation"))
        self.assertTrue(callable(self.dataset.delete_annotation))

    def test_delete_annotation_deletes_annotation_record(self):
        self.dataset.annotate(self.annotation)
        orig_len_annotations = len(self.dataset.annotations)
        self.dataset.delete_annotation(0)
        new_len_annotations = len(self.dataset.annotations)
        self.assertGreater(orig_len_annotations, new_len_annotations)

    def test_delete_annotation_deletes_correct_annotation_record(self):
        self.dataset.annotate(self.annotation)
        self.dataset.annotate(self.annotation)
        annotation_step = self.dataset.annotations[-1]
        self.dataset.delete_annotation(0)
        self.assertIs(annotation_step, self.dataset.annotations[-1])

    def test_dataset_annotate_sets_package_in_sysinfo(self):
        # Fake package name
        self.dataset._package_name = "numpy"
        self.dataset.annotate(self.annotation)
        annotation_record = self.dataset.annotations[0]
        self.assertTrue("numpy" in annotation_record.sysinfo.packages.keys())

    def test_annotate_adds_task(self):
        self.dataset.annotate(self.annotation)
        self.assertNotEqual(self.dataset.tasks, [])

    def test_added_task_has_kind_annotation(self):
        self.dataset.annotate(self.annotation)
        self.assertEqual(self.dataset.tasks[0]["kind"], "annotation")

    def test_added_task_has_annotation_history_record(self):
        self.dataset.annotate(self.annotation)
        self.assertIsInstance(
            self.dataset.tasks[0]["task"],
            aspecd.history.AnnotationHistoryRecord,
        )


class TestDatasetPlotting(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.plotter = plotting.SinglePlotter()

    def test_has_plot_method(self):
        self.assertTrue(hasattr(self.dataset, "plot"))
        self.assertTrue(callable(self.dataset.plot))

    def test_dataset_plot_returns_plotter_object(self):
        plotter_object = plotting.SinglePlotter()
        plot = self.dataset.plot(plotter_object)
        self.assertTrue(isinstance(plot, plotting.Plotter))

    def test_plot_without_plotter_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingPlotterError):
            self.dataset.plot()

    def test_plot_adds_task(self):
        self.dataset.plot(self.plotter)
        self.assertNotEqual(self.dataset.tasks, [])

    def test_added_task_has_kind_representation(self):
        self.dataset.plot(self.plotter)
        self.assertEqual(self.dataset.tasks[0]["kind"], "representation")

    def test_added_task_has_plot_history_record(self):
        self.dataset.plot(self.plotter)
        self.assertIsInstance(
            self.dataset.tasks[0]["task"], aspecd.history.PlotHistoryRecord
        )


class TestDatasetTabulating(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.table = table.Table()

    def test_has_tabulate_method(self):
        self.assertTrue(hasattr(self.dataset, "tabulate"))
        self.assertTrue(callable(self.dataset.tabulate))

    def test_dataset_tabulate_returns_table_object(self):
        tab = self.dataset.tabulate(self.table)
        self.assertTrue(isinstance(tab, table.Table))

    def test_plot_without_table_raises(self):
        with self.assertRaisesRegex(
            TypeError, "tabulate needs a Table object"
        ):
            self.dataset.tabulate()

    def test_tabulate_adds_task(self):
        self.dataset.tabulate(self.table)
        self.assertNotEqual(self.dataset.tasks, [])

    def test_added_task_has_kind_representation(self):
        self.dataset.tabulate(self.table)
        self.assertEqual(self.dataset.tasks[0]["kind"], "representation")

    def test_added_task_has_table_history_record(self):
        self.dataset.tabulate(self.table)
        self.assertIsInstance(
            self.dataset.tasks[0]["task"], aspecd.history.TableHistoryRecord
        )


class TestDatasetRepresentations(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.plotter = plotting.SinglePlotter()

    def test_plot_adds_plot_record_to_representations(self):
        self.dataset.plot(self.plotter)
        self.assertFalse(self.dataset.representations == [])

    def test_added_plot_record_is_plotrecord(self):
        self.dataset.plot(self.plotter)
        self.assertTrue(
            isinstance(
                self.dataset.representations[-1],
                aspecd.history.PlotHistoryRecord,
            )
        )

    def test_added_plot_record_contains_history(self):
        processing_step = processing.SingleProcessingStep()
        self.dataset.process(processing_step)
        self.dataset.plot(self.plotter)
        representation = self.dataset.representations[-1]
        self.assertEqual(
            len(self.dataset.history), len(representation.plot.preprocessing)
        )

    def test_added_plot_record_history_is_deepcopy(self):
        processing_step = processing.SingleProcessingStep()
        self.dataset.process(processing_step)
        self.dataset.plot(self.plotter)
        representation = self.dataset.representations[-1]
        self.assertNotEqual(
            self.dataset.history, representation.plot.preprocessing
        )

    def test_has_delete_representation_method(self):
        self.assertTrue(hasattr(self.dataset, "delete_representation"))
        self.assertTrue(callable(self.dataset.delete_representation))

    def test_delete_representation_deletes_representation_record(self):
        self.dataset.plot(self.plotter)
        orig_len_representations = len(self.dataset.representations)
        self.dataset.delete_representation(0)
        new_len_representations = len(self.dataset.representations)
        self.assertGreater(orig_len_representations, new_len_representations)

    def test_delete_representation_deletes_correct_representation_record(
        self,
    ):
        self.dataset.plot(self.plotter)
        self.dataset.plot(self.plotter)
        representation = self.dataset.representations[-1]
        self.dataset.delete_representation(0)
        self.assertIs(representation, self.dataset.representations[-1])


class TestDatasetImporting(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.importer = io.DatasetImporter()

    def test_import_without_importer_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingImporterError):
            self.dataset.import_from()


class TestDatasetExporting(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.exporter = io.DatasetExporter()

    def test_export_without_exporter_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingExporterError):
            self.dataset.export_to()

    def test_export_adds_task(self):
        self.dataset.export_to(self.exporter)
        self.assertNotEqual(self.dataset.tasks, [])

    def test_added_task_has_kind_export(self):
        self.dataset.export_to(self.exporter)
        self.assertEqual(self.dataset.tasks[0]["kind"], "export")

    def test_added_task_has_exporter_history_record(self):
        self.dataset.export_to(self.exporter)
        self.assertIsInstance(
            self.dataset.tasks[0]["task"],
            aspecd.history.DatasetExporterHistoryRecord,
        )


class TestDatasetToDict(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()
        self.figure_filename = "foo.pdf"

    def tearDown(self):
        if os.path.exists(self.figure_filename):
            os.remove(self.figure_filename)

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.dataset, "to_dict"))
        self.assertTrue(callable(self.dataset.to_dict))

    def test_to_dict_with_empty_dataset(self):
        self.dataset.to_dict()

    def test_to_dict_with_processing_step(self):
        processing_step = aspecd.processing.SingleProcessingStep()
        self.dataset.process(processing_step)
        self.dataset.to_dict()

    def test_to_dict_with_analysis_step(self):
        analysis_step = aspecd.analysis.SingleAnalysisStep()
        self.dataset.analyse(analysis_step)
        self.dataset.to_dict()

    def test_to_dict_with_annotation(self):
        annotation_step = aspecd.annotation.DatasetAnnotation()
        annotation_step.content = "foo"
        self.dataset.annotate(annotation_step)
        self.dataset.to_dict()

    def test_to_dict_with_representation(self):
        plotter = aspecd.plotting.SinglePlotter()
        self.dataset.plot(plotter)
        self.dataset.to_dict()

    def test_to_dict_with_representation_from_plotter(self):
        plotter = aspecd.plotting.SinglePlotter()
        plotter.plot(self.dataset)
        self.dataset.to_dict()

    def test_to_dict_with_saved_representation(self):
        plotter = aspecd.plotting.SinglePlotter()
        saver = aspecd.plotting.Saver(filename=self.figure_filename)
        plot = self.dataset.plot(plotter)
        saver.plotter = plot
        saver.save()
        self.dataset.to_dict()

    def test_to_dict_includes_origdata(self):
        self.assertIn("_origdata", self.dataset.to_dict())

    def test_to_dict_includes_package_name(self):
        self.assertIn("_package_name", self.dataset.to_dict())

    def test_to_dict_includes_history_pointer(self):
        self.assertIn("_history_pointer", self.dataset.to_dict())

    def test_to_dict_with_device_data(self):
        self.dataset.device_data["foo"] = aspecd.dataset.DeviceData()
        self.assertNotIn("_origdata", self.dataset.to_dict()["device_data"])


class TestDatasetFromDict(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.ExperimentalDataset()

    def test_has_from_dict_method(self):
        self.assertTrue(hasattr(self.dataset, "from_dict"))
        self.assertTrue(callable(self.dataset.from_dict))

    def test_from_dict_sets_id(self):
        dataset_dict = self.dataset.to_dict()
        dataset_dict["id"] = "foo"
        self.dataset.from_dict(dataset_dict)
        self.assertEqual(dataset_dict["id"], self.dataset.id)

    def test_from_dict_sets_data(self):
        dataset_dict = self.dataset.to_dict()
        dataset_dict["data"]["data"] = np.random.rand(10)
        self.dataset.from_dict(dataset_dict)
        np.testing.assert_allclose(
            dataset_dict["data"]["data"], self.dataset.data.data
        )

    def test_from_dict_sets_origdata(self):
        dataset_dict = self.dataset.to_dict()
        dataset_dict["_origdata"]["data"] = np.random.rand(10)
        self.dataset.from_dict(dataset_dict)
        np.testing.assert_allclose(
            dataset_dict["_origdata"]["data"], self.dataset._origdata.data
        )

    def test_from_dict_sets_axes_in_data(self):
        dataset_dict = self.dataset.to_dict()
        dataset_dict["data"]["axes"][0]["label"] = "foo"
        self.dataset.from_dict(dataset_dict)
        self.assertEqual(
            dataset_dict["data"]["axes"][0]["label"],
            self.dataset.data.axes[0].label,
        )

    def test_from_dict_sets_metadata(self):
        dataset_dict = self.dataset.to_dict()
        dataset_dict["metadata"]["sample"]["name"] = "foo"
        self.dataset.from_dict(dataset_dict)
        self.assertEqual(
            dataset_dict["metadata"]["sample"]["name"],
            self.dataset.metadata.sample.name,
        )

    def test_from_dict_sets_history(self):
        processing_step = processing.SingleProcessingStep()
        self.dataset.process(processing_step)
        dataset_dict = self.dataset.to_dict()
        new_dataset = dataset.Dataset()
        new_dataset.from_dict(dataset_dict)
        self.assertDictEqual(
            self.dataset.history[0].to_dict(),
            new_dataset.history[0].to_dict(),
        )

    def test_from_dict_sets_analyses(self):
        analysis_step = aspecd.analysis.SingleAnalysisStep()
        self.dataset.analyse(analysis_step)
        dataset_dict = self.dataset.to_dict()
        new_dataset = dataset.Dataset()
        new_dataset.from_dict(dataset_dict)
        self.assertDictEqual(
            self.dataset.analyses[0].to_dict(),
            new_dataset.analyses[0].to_dict(),
        )

    def test_from_dict_sets_annotations(self):
        annotation_ = aspecd.annotation.DatasetAnnotation()
        annotation_.content = {"foo": "bar"}
        self.dataset.annotate(annotation_)
        dataset_dict = self.dataset.to_dict()
        new_dataset = dataset.Dataset()
        new_dataset.from_dict(dataset_dict)
        self.assertDictEqual(
            self.dataset.annotations[0].to_dict(),
            new_dataset.annotations[0].to_dict(),
        )

    def test_from_dict_sets_representations(self):
        representation = aspecd.plotting.SinglePlotter1D()
        self.dataset.plot(representation)
        dataset_dict = self.dataset.to_dict()
        new_dataset = dataset.Dataset()
        new_dataset.from_dict(dataset_dict)
        self.assertDictEqual(
            self.dataset.representations[0].to_dict(),
            new_dataset.representations[0].to_dict(),
        )

    def test_from_dict_sets_references(self):
        reference_dataset = aspecd.dataset.Dataset()
        reference_dataset.id = "foo"
        self.dataset.add_reference(reference_dataset)
        dataset_dict = self.dataset.to_dict()
        new_dataset = dataset.Dataset()
        new_dataset.from_dict(dataset_dict)
        self.assertDictEqual(
            self.dataset.references[0].to_dict(),
            new_dataset.references[0].to_dict(),
        )

    def test_from_dict_sets_tasks(self):
        processing_step = processing.SingleProcessingStep()
        self.dataset.process(processing_step)
        dataset_dict = self.dataset.to_dict()
        new_dataset = dataset.Dataset()
        new_dataset.from_dict(dataset_dict)
        self.assertDictEqual(
            self.dataset.tasks[0]["task"].to_dict(),
            new_dataset.tasks[0]["task"].to_dict(),
        )

    def test_from_dict_sets_package_name(self):
        dataset_dict = self.dataset.to_dict()
        dataset_dict["_package_name"] = "foo"
        new_dataset = dataset.Dataset()
        new_dataset.from_dict(dataset_dict)
        self.assertEqual(
            dataset_dict["_package_name"], new_dataset.package_name
        )

    def test_from_dict_sets_history_pointer(self):
        dataset_dict = self.dataset.to_dict()
        dataset_dict["_history_pointer"] = 42
        new_dataset = dataset.Dataset()
        new_dataset.from_dict(dataset_dict)
        self.assertEqual(
            dataset_dict["_history_pointer"], new_dataset._history_pointer
        )


class TestDatasetReferences(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.Dataset()

    def test_has_add_reference_method(self):
        self.assertTrue(hasattr(self.dataset, "add_reference"))
        self.assertTrue(callable(self.dataset.add_reference))

    def test_add_reference_without_dataset_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.dataset.add_reference()

    def test_add_reference_adds_dataset_to_references(self):
        new_dataset = aspecd.dataset.CalculatedDataset()
        self.dataset.add_reference(new_dataset)
        self.assertTrue(self.dataset.references)

    def test_add_reference_adds_dataset_references_to_references(self):
        new_dataset = aspecd.dataset.CalculatedDataset()
        self.dataset.add_reference(new_dataset)
        self.assertTrue(
            isinstance(
                self.dataset.references[0], aspecd.dataset.DatasetReference
            )
        )

    def test_has_remove_reference_method(self):
        self.assertTrue(hasattr(self.dataset, "remove_reference"))
        self.assertTrue(callable(self.dataset.remove_reference))

    def test_remove_reference_without_id_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.dataset.remove_reference()

    def test_remove_reference_removes_dataset_from_references(self):
        new_dataset = aspecd.dataset.CalculatedDataset()
        new_dataset.id = "foo"
        self.dataset.add_reference(new_dataset)
        self.dataset.remove_reference(dataset_id=new_dataset.id)
        self.assertFalse(self.dataset.references)

    def test_remove_reference_with_wrong_id_doesnt_remove_dataset(self):
        new_dataset = aspecd.dataset.CalculatedDataset()
        new_dataset.id = "foo"
        self.dataset.add_reference(new_dataset)
        self.dataset.remove_reference(dataset_id="bar")
        self.assertTrue(self.dataset.references)


class TestExperimentalDataset(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.ExperimentalDataset()

    def test_instantiate_class(self):
        pass

    def test_metadata_is_ExperimentalDatasetMetadata(self):
        self.assertTrue(
            isinstance(
                self.dataset.metadata,
                aspecd.metadata.ExperimentalDatasetMetadata,
            )
        )


class TestCalculatedDataset(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.CalculatedDataset()

    def test_instantiate_class(self):
        pass

    def test_metadata_is_CalculatedDatasetMetadata(self):
        self.assertTrue(
            isinstance(
                self.dataset.metadata,
                aspecd.metadata.CalculatedDatasetMetadata,
            )
        )

    def test_data_is_calculated(self):
        self.assertTrue(self.dataset.data.calculated)

    def test_origdata_is_calculated(self):
        self.assertTrue(self.dataset._origdata.calculated)


class TestDatasetReference(unittest.TestCase):
    def setUp(self):
        self.reference = dataset.DatasetReference()
        self.dataset = dataset.Dataset()

    def test_instantiate_class(self):
        pass

    def test_has_type_property(self):
        self.assertTrue(hasattr(self.reference, "type"))

    def test_has_id_property(self):
        self.assertTrue(hasattr(self.reference, "id"))

    def test_has_history_property(self):
        self.assertTrue(hasattr(self.reference, "history"))

    def test_has_from_dataset_method(self):
        self.assertTrue(hasattr(self.reference, "from_dataset"))
        self.assertTrue(callable(self.reference.from_dataset))

    def test_from_dataset_without_dataset_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.reference.from_dataset()

    def test_from_dataset_sets_type(self):
        self.reference.from_dataset(self.dataset)
        self.assertEqual(
            aspecd.utils.full_class_name(self.dataset), self.reference.type
        )

    def test_from_dataset_sets_id(self):
        self.dataset.source = "foo"
        self.reference.from_dataset(self.dataset)
        self.assertEqual(self.dataset.id, self.reference.id)

    def test_from_dataset_sets_history(self):
        processing_step = aspecd.processing.SingleProcessingStep()
        self.dataset.process(processing_step)
        self.reference.from_dataset(self.dataset)
        self.assertTrue(self.reference.history)

    def test_from_dataset_copies_history(self):
        processing_step = aspecd.processing.SingleProcessingStep()
        self.dataset.process(processing_step)
        self.reference.from_dataset(self.dataset)
        self.assertIsNot(self.dataset.history, self.reference.history)

    def test_has_to_dataset_method(self):
        self.assertTrue(hasattr(self.reference, "to_dataset"))
        self.assertTrue(callable(self.reference.to_dataset))

    def test_to_dataset_without_type_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.reference.to_dataset()

    def test_to_dataset_returns_dataset(self):
        self.reference.type = aspecd.utils.full_class_name(self.dataset)
        dataset_ = self.reference.to_dataset()
        self.assertTrue(isinstance(dataset_, aspecd.dataset.Dataset))

    def test_to_dataset_returns_dataset_of_correct_type(self):
        original_dataset = aspecd.dataset.CalculatedDataset()
        self.reference.from_dataset(original_dataset)
        new_dataset = self.reference.to_dataset()
        self.assertTrue(
            isinstance(new_dataset, aspecd.dataset.CalculatedDataset)
        )

    def test_to_dataset_sets_dataset_id(self):
        self.reference.type = aspecd.utils.full_class_name(self.dataset)
        self.reference.id = "foo"
        dataset_ = self.reference.to_dataset()
        self.assertEqual(self.reference.id, dataset_.id)

    def test_to_dataset_applies_history(self):
        processing_step = aspecd.processing.SingleProcessingStep()
        self.dataset.process(processing_step)
        self.reference.from_dataset(self.dataset)
        dataset_ = self.reference.to_dataset()
        self.assertTrue(dataset_.history)

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.reference, "to_dict"))
        self.assertTrue(callable(self.reference.to_dict))

    def test_from_dict(self):
        orig_dict = self.reference.to_dict()
        orig_dict["id"] = "foo"
        new_dataset_record = aspecd.dataset.DatasetReference()
        new_dataset_record.from_dict(orig_dict)
        self.assertDictEqual(orig_dict, new_dataset_record.to_dict())


class TestDatasetFactory(unittest.TestCase):
    def setUp(self):
        self.factory = dataset.DatasetFactory()
        self.source = "/foo"
        self.filename = "foo"

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_instantiate_class(self):
        pass

    def test_has_importer_factory_property(self):
        self.assertTrue(hasattr(self.factory, "importer_factory"))

    def test_get_dataset_returns_dataset(self):
        self.factory.importer_factory = io.DatasetImporterFactory()
        dataset_ = self.factory.get_dataset(source=self.source)
        self.assertTrue(isinstance(dataset_, dataset.Dataset))

    def test_get_dataset_without_source_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingSourceError):
            self.factory.get_dataset()

    def test_get_dataset_without_importer_factory_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingImporterFactoryError):
            self.factory.get_dataset(source=self.source)

    def test_get_dataset_sets_id_from_source_in_dataset(self):
        self.factory.importer_factory = io.DatasetImporterFactory()
        dataset_ = self.factory.get_dataset(source=self.source)
        self.assertEqual(self.source, dataset_.id)

    def test_get_dataset_with_importer(self):
        data = np.random.random(1)
        np.savetxt(self.filename, data)
        self.factory.importer_factory = io.DatasetImporterFactory()
        dataset_ = self.factory.get_dataset(
            source=self.filename, importer="TxtImporter"
        )
        self.assertEqual(data, dataset_.data.data)

    def test_get_dataset_with_importer_and_parameters(self):
        parameters = {"skiprows": 1}
        data = np.random.random(2)
        np.savetxt(self.filename, data)
        self.factory.importer_factory = io.DatasetImporterFactory()
        dataset_ = self.factory.get_dataset(
            source=self.filename,
            importer="TxtImporter",
            parameters=parameters,
        )
        self.assertEqual(data[1], dataset_.data.data)


class TestData(unittest.TestCase):
    def setUp(self):
        self.data = dataset.Data()

    def test_instantiate_class(self):
        pass

    def test_has_data_property(self):
        self.assertTrue(hasattr(self.data, "data"))

    def test_data_is_ndarray(self):
        self.assertTrue(isinstance(self.data.data, np.ndarray))

    def test_has_axes_property(self):
        self.assertTrue(hasattr(self.data, "axes"))

    def test_axes_is_list(self):
        self.assertTrue(isinstance(self.data.axes, list))

    def test_axes_have_right_count_for_1d_data(self):
        self.data.data = np.zeros(0)
        self.assertEqual(len(self.data.axes), 2)

    def test_axes_have_right_count_for_2d_data(self):
        self.data.data = np.zeros([0, 0])
        self.assertEqual(len(self.data.axes), 3)

    def test_has_calculated_property(self):
        self.assertTrue(hasattr(self.data, "calculated"))

    def test_calculated_is_boolean(self):
        self.assertTrue(isinstance(self.data.calculated, bool))

    def test_modify_data_with_same_dimension_does_not_change_axes(self):
        data = np.zeros(5)
        axis_values = np.arange(len(data))
        self.data.data = data
        self.data.axes[0].values = axis_values
        self.data.data = data
        self.assertTrue(np.allclose(self.data.axes[0].values, axis_values))

    def test_modify_data_with_different_dimensions_keeps_axes_metadata(self):
        old_data = np.zeros([5, 1])
        new_data = np.zeros([5, 2])
        axis_quantity = "foobar"
        self.data.data = old_data
        self.data.axes[0].values = np.arange(len(old_data))
        self.data.axes[0].quantity = axis_quantity
        self.data.data = new_data
        self.assertTrue(self.data.axes[0].quantity, axis_quantity)

    def test_modify_data_with_different_dimensions_adjusts_axes_values(self):
        old_data = np.zeros([5, 1])
        new_data = np.zeros([7, 1])
        self.data.data = old_data
        self.data.axes[0].values = np.arange(len(old_data))
        self.data.data = new_data
        self.assertEqual(new_data.size, self.data.axes[0].values.size)

    def test_modify_data_with_higher_dimension_adds_axis(self):
        old_data = np.random.random([5, 1])
        new_data = np.random.random([4, 3, 2])
        self.data.data = old_data
        self.data.data = new_data
        self.assertEqual(4, len(self.data.axes))

    def test_setting_data_adjusts_axes_values(self):
        new_data = np.zeros([6, 1])
        self.data.data = new_data
        self.assertEqual(new_data.size, self.data.axes[0].values.size)

    def test_setting_data_adjusts_axes_values_for_2D_dataset(self):
        self.data.data = np.zeros([6, 3])
        shape = self.data.data.shape
        self.assertEqual(shape[0], len(self.data.axes[0].values))
        self.assertEqual(shape[1], len(self.data.axes[1].values))

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.data, "to_dict"))
        self.assertTrue(callable(self.data.to_dict))

    def test_to_dict_includes_data(self):
        dict_ = self.data.to_dict()
        self.assertIn("data", dict_)

    def test_to_dict_includes_axes(self):
        dict_ = self.data.to_dict()
        self.assertIn("axes", dict_)

    def test_has_from_dict_method(self):
        self.assertTrue(hasattr(self.data, "from_dict"))
        self.assertTrue(callable(self.data.from_dict))

    def test_from_dict_sets_data(self):
        orig_dict = self.data.to_dict()
        orig_dict["data"] = np.random.rand(10)
        self.data.from_dict(orig_dict)
        np.testing.assert_allclose(orig_dict["data"], self.data.data)

    def test_from_dict_sets_axes(self):
        orig_dict = self.data.to_dict()
        orig_dict["data"] = np.random.rand(10)
        orig_dict["axes"][0]["values"] = np.arange(10)
        orig_dict["axes"][0]["label"] = "foo"
        orig_dict["axes"][1]["label"] = "bar"
        orig_dict["axes"][0]["index"] = [
            "" for _ in orig_dict["axes"][0]["values"]
        ]
        self.data.from_dict(orig_dict)
        self.assertDictEqual(orig_dict, self.data.to_dict())
        # np.testing.assert_allclose(orig_dict["data"], self.data.data)


class TestAxisSetupInConstructor(unittest.TestCase):
    def setUp(self):
        self.data = np.zeros(0)
        self.axes = [dataset.Axis(), dataset.Axis()]
        self.calculated = True

    def test_set_data_in_constructor(self):
        data_obj = dataset.Data(data=self.data)
        self.assertEqual(data_obj.data.tolist(), self.data.tolist())

    def test_set_axes_in_constructor(self):
        data_obj = dataset.Data(axes=self.axes)
        self.assertEqual(data_obj.axes, self.axes)

    def test_set_calculated_in_constructor(self):
        data_obj = dataset.Data(calculated=self.calculated)
        self.assertEqual(data_obj.calculated, self.calculated)

    def test_setting_too_many_axes_raises(self):
        axes = self.axes
        axes.append(dataset.Axis())
        with self.assertRaises(aspecd.exceptions.AxesCountError):
            dataset.Data(self.data, axes)

    def test_axes_values_dimensions_are_consistent_with_empty_1D_data(self):
        data_obj = dataset.Data(self.data, self.axes)
        self.assertEqual(len(data_obj.axes[0].values), 0)

    def test_axes_values_dimensions_are_consistent_with_empty_2D_data(self):
        tmp_data = np.zeros([0, 0])
        data_obj = dataset.Data(tmp_data)
        self.assertEqual(len(data_obj.axes[0].values), 0)
        self.assertEqual(len(data_obj.axes[1].values), 0)

    def test_axes_values_dimensions_are_consistent_with_nonempty_1D_data(
        self,
    ):
        len_data = 5
        tmp_data = np.zeros(len_data)
        tmp_axes = [dataset.Axis(), dataset.Axis()]
        tmp_axes[0].values = np.zeros(len_data)
        data_obj = dataset.Data(tmp_data, tmp_axes)
        self.assertEqual(len(data_obj.axes[0].values), len_data)

    def test_axes_values_dimensions_are_consistent_with_nonempty_2D_data(
        self,
    ):
        len_data = [5, 3]
        tmp_data = np.zeros(len_data)
        tmp_axes = [dataset.Axis(), dataset.Axis(), dataset.Axis()]
        tmp_axes[0].values = np.zeros(len_data[0])
        tmp_axes[1].values = np.zeros(len_data[1])
        data_obj = dataset.Data(tmp_data, tmp_axes)
        self.assertEqual(len(data_obj.axes[0].values), len_data[0])
        self.assertEqual(len(data_obj.axes[1].values), len_data[1])

    def test_wrong_axes_values_dimensions_with_nonempty_1D_data_raises(self):
        len_data = 5
        tmp_data = np.zeros(len_data)
        with self.assertRaises(
            aspecd.exceptions.AxesValuesInconsistentWithDataError
        ):
            dataset.Data(tmp_data, self.axes)

    def test_wrong_axes_values_dimensions_with_nonempty_2D_data_raises(self):
        len_data = [5, 3]
        tmp_data = np.zeros(len_data)
        with self.assertRaises(
            aspecd.exceptions.AxesValuesInconsistentWithDataError
        ):
            dataset.Data(tmp_data, self.axes)

    def test_set_wrong_axes_dimensions_with_nonempty_1D_data_raises(self):
        len_data = 5
        tmp_data = np.zeros(len_data)
        tmp_axis = dataset.Axis()
        tmp_axis.values = np.zeros(2 * len_data)
        tmp_axes = [tmp_axis, dataset.Axis()]
        with self.assertRaises(
            aspecd.exceptions.AxesValuesInconsistentWithDataError
        ):
            dataset.Data(tmp_data, tmp_axes)

    def test_set_wrong_axes_dimensions_with_nonempty_2D_data_raises(self):
        len_data = [5, 3]
        tmp_data = np.zeros(len_data)
        tmp_axis1 = dataset.Axis()
        tmp_axis1.values = np.zeros(2 * len_data[0])
        tmp_axis2 = dataset.Axis()
        tmp_axis2.values = np.zeros(2 * len_data[1])
        tmp_axes = [tmp_axis1, tmp_axis2, dataset.Axis()]
        with self.assertRaises(
            aspecd.exceptions.AxesValuesInconsistentWithDataError
        ):
            dataset.Data(tmp_data, tmp_axes)


class TestAxis(unittest.TestCase):
    def setUp(self):
        self.axis = dataset.Axis()

    def test_instantiate_class(self):
        pass

    def test_has_values_property(self):
        self.assertTrue(hasattr(self.axis, "values"))

    def test_values_is_ndarray(self):
        self.assertTrue(isinstance(self.axis.values, np.ndarray))

    def test_values_is_1d(self):
        self.assertTrue(self.axis.values.ndim, 1)

    def test_has_quantity_property(self):
        self.assertTrue(hasattr(self.axis, "quantity"))

    def test_quantity_is_string(self):
        self.assertTrue(isinstance(self.axis.quantity, str))

    def test_has_symbol_property(self):
        self.assertTrue(hasattr(self.axis, "symbol"))

    def test_symbol_is_string(self):
        self.assertTrue(isinstance(self.axis.symbol, str))

    def test_has_unit_property(self):
        self.assertTrue(hasattr(self.axis, "unit"))

    def test_unit_is_string(self):
        self.assertTrue(isinstance(self.axis.unit, str))

    def test_has_label_property(self):
        self.assertTrue(hasattr(self.axis, "label"))

    def test_label_is_string(self):
        self.assertTrue(isinstance(self.axis.label, str))

    def test_has_equidistant_property(self):
        self.assertTrue(hasattr(self.axis, "equidistant"))

    def test_equidistant_is_none_by_default(self):
        self.assertEqual(self.axis.equidistant, None)

    def test_equidistant_is_true_for_equidistant_axes(self):
        self.axis.values = np.linspace(330, 340, num=1024, endpoint=True)
        self.assertTrue(self.axis.equidistant)

    def test_equidistant_is_false_for_nonequidistant_axes(self):
        self.axis.values = np.asarray([0, 1, 2, 4, 8])
        self.assertFalse(self.axis.equidistant)

    def test_has_index_property(self):
        self.assertTrue(hasattr(self.axis, "index"))

    def test_index_is_list(self):
        self.assertTrue(isinstance(self.axis.index, list))

    def test_index_is_same_length_as_values(self):
        length = 5
        self.axis.values = np.zeros(length)
        self.assertEqual(length, len(self.axis.index))

    def test_setting_index_with_incompatible_length_to_values_raises(self):
        self.axis.values = np.zeros(5)
        indices = ["foo", "bar"]
        with self.assertRaisesRegex(
            IndexError, "index and values need to be " "of same length"
        ):
            self.axis.index = indices

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.axis, "to_dict"))
        self.assertTrue(callable(self.axis.to_dict))

    def test_to_dict_includes_values(self):
        dict_ = self.axis.to_dict()
        self.assertIn("values", dict_)

    def test_to_dict_includes_index(self):
        dict_ = self.axis.to_dict()
        self.assertIn("index", dict_)

    def test_has_from_dict_method(self):
        self.assertTrue(hasattr(self.axis, "from_dict"))
        self.assertTrue(callable(self.axis.from_dict))

    def test_from_dict_sets_string_values(self):
        orig_dict = self.axis.to_dict()
        orig_dict["label"] = "foo"
        self.axis.from_dict(orig_dict)
        new_dict = self.axis.to_dict()
        self.assertDictEqual(orig_dict, new_dict)

    def test_from_dict_sets_numeric_values(self):
        orig_dict = self.axis.to_dict()
        orig_dict["values"] = np.arange(10)
        orig_dict["index"] = ["" for _ in orig_dict["values"]]
        self.axis.from_dict(orig_dict)
        new_dict = self.axis.to_dict()
        self.assertDictEqual(orig_dict, new_dict)

    def test_set_values(self):
        self.axis.values = np.zeros(0)

    def test_set_wrong_type_for_values_fails(self):
        with self.assertRaisesRegex(ValueError, "Wrong type: expected"):
            self.axis.values = "foo"

    def test_set_multidimensional_values_fails(self):
        with self.assertRaisesRegex(
            IndexError, "Values need to be " "one-dimensional"
        ):
            self.axis.values = np.zeros([0, 0])


class TestDeviceData(unittest.TestCase):
    def setUp(self):
        self.device_data = dataset.DeviceData()

    def test_instantiate_class(self):
        pass

    def test_has_data_property(self):
        self.assertTrue(hasattr(self.device_data, "data"))

    def test_data_is_ndarray(self):
        self.assertTrue(isinstance(self.device_data.data, np.ndarray))

    def test_has_axes_property(self):
        self.assertTrue(hasattr(self.device_data, "axes"))

    def test_axes_is_list(self):
        self.assertTrue(isinstance(self.device_data.axes, list))

    def test_has_metadata_property(self):
        self.assertTrue(hasattr(self.device_data, "metadata"))

    def test_metadata_property_is_device_metadata(self):
        self.assertIsInstance(
            self.device_data.metadata, aspecd.metadata.Device
        )

    def test_calculated_is_false(self):
        self.assertFalse(self.device_data.calculated)



"""Tests for analysis."""

import unittest

import numpy as np
import scipy.signal

import aspecd.analysis
import aspecd.dataset
import aspecd.exceptions
import aspecd.history
import aspecd.processing


class TestAnalysisStep(unittest.TestCase):
    def setUp(self):
        self.analysisstep = aspecd.analysis.AnalysisStep()

    def test_instantiate_class(self):
        pass

    def test_has_name_property(self):
        self.assertTrue(hasattr(self.analysisstep, "name"))

    def test_name_property_equals_class_name(self):
        class_name = "aspecd.analysis.AnalysisStep"
        self.assertEqual(self.analysisstep.name, class_name)

    def test_has_parameters_property(self):
        self.assertTrue(hasattr(self.analysisstep, "parameters"))

    def test_parameters_property_is_dict(self):
        self.assertTrue(isinstance(self.analysisstep.parameters, dict))

    def test_has_result_property(self):
        self.assertTrue(hasattr(self.analysisstep, "result"))

    def test_has_description_property(self):
        self.assertTrue(hasattr(self.analysisstep, "description"))

    def test_description_property_is_string(self):
        self.assertTrue(isinstance(self.analysisstep.description, str))

    def test_has_comment_property(self):
        self.assertTrue(hasattr(self.analysisstep, "comment"))

    def test_description_comment_is_string(self):
        self.assertTrue(isinstance(self.analysisstep.comment, str))

    def test_has_references_property(self):
        self.assertTrue(hasattr(self.analysisstep, "references"))

    def test_description_references_is_list(self):
        self.assertTrue(isinstance(self.analysisstep.references, list))

    def test_has_index_property(self):
        self.assertTrue(hasattr(self.analysisstep, "index"))

    def test_index_is_list(self):
        self.assertTrue(isinstance(self.analysisstep.index, list))

    def test_has_analyse_method(self):
        self.assertTrue(hasattr(self.analysisstep, "analyse"))
        self.assertTrue(callable(self.analysisstep.analyse))

    def test_has_analyze_method(self):
        self.assertTrue(hasattr(self.analysisstep, "analyze"))
        self.assertTrue(callable(self.analysisstep.analyze))

    def test_has_applicable_method(self):
        self.assertTrue(hasattr(self.analysisstep, "applicable"))
        self.assertTrue(callable(self.analysisstep.applicable))

    def test_create_dataset_creates_calculated_dataset(self):
        self.assertIsInstance(
            self.analysisstep.create_dataset(),
            aspecd.dataset.CalculatedDataset,
        )

    def test_created_dataset_has_correct_type_in_metadata(self):
        dataset = self.analysisstep.create_dataset()
        self.assertEqual(
            self.analysisstep.name, dataset.metadata.calculation.type
        )

    def test_created_dataset_contains_parameters_in_metadata(self):
        self.analysisstep.parameters["foo"] = "bar"
        dataset = self.analysisstep.create_dataset()
        self.assertDictEqual(
            self.analysisstep.parameters,
            dataset.metadata.calculation.parameters,
        )

    def test_create_dataset_with_dataset_type_returns_correct_type(self):
        self.analysisstep.dataset_type = "aspecd.dataset.CalculatedDataset"
        self.assertIsInstance(
            self.analysisstep.create_dataset(),
            aspecd.dataset.CalculatedDataset,
        )

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.analysisstep, "to_dict"))
        self.assertTrue(callable(self.analysisstep.to_dict))

    def test_to_dict_does_not_contain_certain_keys(self):
        for key in ["name", "description", "references", "result"]:
            with self.subTest(key=key):
                self.assertNotIn(key, self.analysisstep.to_dict())


class TestSingleAnalysisStep(unittest.TestCase):
    def setUp(self):
        self.analysisstep = aspecd.analysis.SingleAnalysisStep()

    def test_instantiate_class(self):
        pass

    def test_has_preprocessing_property(self):
        self.assertTrue(hasattr(self.analysisstep, "preprocessing"))

    def test_preprocessing_is_list(self):
        self.assertTrue(isinstance(self.analysisstep.preprocessing, list))

    def test_to_dict_does_not_contain_certain_keys(self):
        for key in ["dataset"]:
            with self.subTest(key=key):
                self.assertNotIn(key, self.analysisstep.to_dict())

    def test_analyse_with_dataset(self):
        test_dataset = aspecd.dataset.Dataset()
        self.analysisstep.analyse(test_dataset)
        self.assertGreater(len(test_dataset.analyses), 0)

    def test_analyse_without_argument_and_with_dataset(self):
        self.analysisstep.dataset = aspecd.dataset.Dataset()
        self.analysisstep.analyse()
        self.assertGreater(len(self.analysisstep.dataset.analyses), 0)

    def test_analyse_without_analysisstep_nor_dataset_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.analysisstep.analyse()

    def test_analyse_returns_dataset(self):
        test_dataset = self.analysisstep.analyse(aspecd.dataset.Dataset())
        self.assertTrue(isinstance(test_dataset, aspecd.dataset.Dataset))

    def test_has_create_history_record_method(self):
        self.assertTrue(hasattr(self.analysisstep, "create_history_record"))
        self.assertTrue(callable(self.analysisstep.create_history_record))

    def test_create_history_record_returns_history_record(self):
        self.analysisstep.dataset = aspecd.dataset.Dataset()
        history_record = self.analysisstep.create_history_record()
        self.assertTrue(
            isinstance(history_record, aspecd.history.AnalysisHistoryRecord)
        )

    def test_analyse_checks_applicability(self):
        class MyAnalysisStep(aspecd.analysis.SingleAnalysisStep):
            @staticmethod
            def applicable(dataset):
                return False

        dataset_ = aspecd.dataset.Dataset()
        analysis = MyAnalysisStep()
        with self.assertRaises(aspecd.exceptions.NotApplicableToDatasetError):
            dataset_.analyse(analysis)

    def test_analyse_checks_applicability_prints_helpful_message(self):
        class MyAnalysisStep(aspecd.analysis.SingleAnalysisStep):
            @staticmethod
            def applicable(dataset):
                return False

        dataset_ = aspecd.dataset.Dataset()
        dataset_.id = "foo"
        analysis = MyAnalysisStep()
        message = "MyAnalysisStep not applicable to dataset with id foo"
        with self.assertRaisesRegex(
            aspecd.exceptions.NotApplicableToDatasetError, message
        ):
            dataset_.analyse(analysis)

    def test_dataset_contains_parameters_obtained_during_analysis(self):
        class MyAnalysisStep(aspecd.analysis.SingleAnalysisStep):
            def _perform_task(self):
                self.result = self.create_dataset()
                self.parameters["foo"] = "bar"

        dataset_ = aspecd.dataset.Dataset()
        analysis = MyAnalysisStep()
        new_analysis = dataset_.analyse(analysis)
        self.assertIn(
            "foo", new_analysis.result.metadata.calculation.parameters
        )

    def test_returned_dataset_contains_correct_origdata(self):
        class MyAnalysisStep(aspecd.analysis.SingleAnalysisStep):
            def _perform_task(self):
                self.result = self.create_dataset()
                self.result.data.data = np.random.random(5)

        dataset_ = aspecd.dataset.Dataset()
        analysis = MyAnalysisStep()
        new_analysis = dataset_.analyse(analysis)
        new_dataset = new_analysis.result
        self.assertListEqual(
            list(new_dataset.data.data), list(new_dataset._origdata.data)
        )


class TestMultiAnalysisStep(unittest.TestCase):
    def setUp(self):
        self.analysisstep = aspecd.analysis.MultiAnalysisStep()

    def test_instantiate_class(self):
        pass

    def test_to_dict_does_not_contain_certain_keys(self):
        for key in ["datasets"]:
            with self.subTest(key=key):
                self.assertNotIn(key, self.analysisstep.to_dict())

    def test_has_datasets_property(self):
        self.assertTrue(hasattr(self.analysisstep, "datasets"))

    def test_datasets_property_is_list(self):
        self.assertTrue(isinstance(self.analysisstep.datasets, list))

    def test_analyse_without_datasets_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.analysisstep.analyse()

    def test_analyse_with_datasets(self):
        self.analysisstep.datasets.append(aspecd.dataset.Dataset())
        self.analysisstep.analyse()

    def test_analyse_checks_applicability(self):
        dataset1 = aspecd.dataset.Dataset()
        dataset1.data.data = np.random.random([5, 5])
        dataset2 = aspecd.dataset.Dataset()
        dataset2.data.data = np.random.random(5)

        class MyAnalysisStep(aspecd.analysis.MultiAnalysisStep):
            @staticmethod
            def applicable(dataset):
                return len(dataset.data.axes) == 2

        analysis_step = MyAnalysisStep()
        analysis_step.datasets.append(dataset1)
        analysis_step.datasets.append(dataset2)
        with self.assertRaises(aspecd.exceptions.NotApplicableToDatasetError):
            analysis_step.analyse()

    def test_analyse_checks_applicability_prints_helpful_message(self):
        dataset1 = aspecd.dataset.Dataset()
        dataset1.data.data = np.random.random([5, 5])
        dataset1.id = "foo"
        dataset2 = aspecd.dataset.Dataset()
        dataset2.data.data = np.random.random(5)
        dataset2.id = "bar"

        class MyAnalysisStep(aspecd.analysis.MultiAnalysisStep):
            @staticmethod
            def applicable(dataset):
                return len(dataset.data.axes) == 2

        analysis_step = MyAnalysisStep()
        analysis_step.datasets.append(dataset1)
        analysis_step.datasets.append(dataset2)
        message = "MyAnalysisStep not applicable to dataset with id foo"
        with self.assertRaisesRegex(
            aspecd.exceptions.NotApplicableToDatasetError, message
        ):
            analysis_step.analyse()

    def test_returned_dataset_contains_correct_origdata(self):
        class MyAnalysisStep(aspecd.analysis.MultiAnalysisStep):
            def _perform_task(self):
                self.result = self.create_dataset()
                self.result.data.data = np.random.random(5)

        analysis_step = MyAnalysisStep()
        analysis_step.datasets.append(aspecd.dataset.Dataset())
        analysis_step.analyse()
        new_dataset = analysis_step.result
        self.assertListEqual(
            list(new_dataset.data.data), list(new_dataset._origdata.data)
        )


class TestAggregatedAnalysisStep(unittest.TestCase):
    def setUp(self):
        self.analysis = aspecd.analysis.AggregatedAnalysisStep()
        self.analysis.datasets.append(aspecd.dataset.Dataset())
        self.analysis.analysis_step = "BasicCharacteristics"
        self.analysis.parameters["kind"] = "min"
        self.analysis.datasets[0].data.data = np.zeros(5)
        self.analysis.datasets[0].label = "foo"

    def test_instantiate_class(self):
        pass

    def test_has_datasets_property(self):
        self.assertTrue(hasattr(self.analysis, "datasets"))

    def test_datasets_property_is_list(self):
        self.assertTrue(isinstance(self.analysis.datasets, list))

    def test_has_analysis_step_property(self):
        self.assertTrue(hasattr(self.analysis, "analysis_step"))

    def test_has_appropriate_description(self):
        self.assertIn(
            "Aggregated analysis step for multiple datasets",
            self.analysis.description,
        )

    def test_to_dict_does_not_contain_certain_keys(self):
        for key in ["datasets", "result"]:
            with self.subTest(key=key):
                self.assertNotIn(key, self.analysis.to_dict())

    def test_analyse_without_datasets_raises(self):
        self.analysis.datasets = []
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.analysis.analyse()

    def test_analyse_without_analysis_step_raises(self):
        self.analysis.analysis_step = ""
        with self.assertRaises(aspecd.exceptions.MissingAnalysisStepError):
            self.analysis.analyse()

    def test_analyse_with_analysis_step_returning_dataset_raises(self):
        self.analysis.analysis_step = "PowerDensitySpectrum"
        with self.assertRaisesRegex(
            ValueError, "Analysis step .* returns dataset"
        ):
            self.analysis.analyse()

    def test_analyse_returns_calculated_dataset_in_result(self):
        self.analysis.analyse()
        self.assertIsInstance(
            self.analysis.result, aspecd.dataset.CalculatedDataset
        )

    def test_result_has_calculation_type_set_correctly(self):
        self.analysis.analyse()
        self.assertEqual(
            "aspecd.analysis.AggregatedAnalysisStep",
            self.analysis.result.metadata.calculation.type,
        )

    def test_analyse_sets_correct_value_in_result(self):
        self.analysis.analyse()
        self.assertEqual(
            min(self.analysis.datasets[0].data.data),
            self.analysis.result.data.data[0],
        )

    def test_analyse_with_full_class_name(self):
        self.analysis.analysis_step = "aspecd.analysis.BasicCharacteristics"
        self.analysis.analyse()
        self.assertEqual(
            min(self.analysis.datasets[0].data.data),
            self.analysis.result.data.data[0],
        )

    def test_analyse_sets_index_in_result(self):
        self.analysis.analyse()
        self.assertEqual(
            self.analysis.datasets[0].label,
            self.analysis.result.data.axes[0].index[0],
        )

    def test_analyse_with_multiple_datasets_sets_correct_values_in_result(
        self,
    ):
        self.analysis.datasets.append(aspecd.dataset.Dataset())
        self.analysis.datasets[1].data.data = np.ones(5)
        self.analysis.datasets[1].label = "bar"
        self.analysis.analyse()
        self.assertEqual(
            min(self.analysis.datasets[0].data.data),
            self.analysis.result.data.data[0],
        )
        self.assertEqual(
            min(self.analysis.datasets[1].data.data),
            self.analysis.result.data.data[1],
        )

    def test_analyse_with_multiple_datasets_sets_index_in_result(self):
        self.analysis.datasets.append(aspecd.dataset.Dataset())
        self.analysis.datasets[1].data.data = np.ones(5)
        self.analysis.datasets[1].label = "bar"
        self.analysis.analyse()
        self.assertEqual(
            self.analysis.datasets[1].label,
            self.analysis.result.data.axes[0].index[1],
        )

    def test_analyse_with_list_return_sets_correct_value_in_result(self):
        self.analysis.parameters["output"] = "indices"
        self.analysis.datasets[0].data.data = np.zeros([5, 5])
        self.analysis.analyse()
        self.assertEqual(
            self.analysis.datasets[0].data.data.ndim,
            len(self.analysis.result.data.data[0, :]),
        )

    def test_analyse_with_list_return_and_multiple_datasets(self):
        self.analysis.parameters["output"] = "indices"
        self.analysis.datasets[0].data.data = np.zeros([5, 5])
        self.analysis.datasets.append(aspecd.dataset.Dataset())
        self.analysis.datasets[1].data.data = np.ones([3, 3])
        self.analysis.datasets.append(aspecd.dataset.Dataset())
        self.analysis.datasets[2].data.data = np.random.random([4, 4])
        self.analysis.analyse()
        self.assertEqual((3, 2), self.analysis.result.data.data.shape)

    def test_returned_dataset_contains_correct_origdata(self):
        self.analysis.analyse()
        self.assertListEqual(
            list(self.analysis.result.data.data),
            list(self.analysis.result._origdata.data),
        )


class TestPreprocessing(unittest.TestCase):
    def setUp(self):
        self.analysisstep = aspecd.analysis.SingleAnalysisStep()
        self.processingstep = aspecd.processing.SingleProcessingStep()

    def test_has_add_preprocessing_step_method(self):
        self.assertTrue(hasattr(self.analysisstep, "add_preprocessing_step"))
        self.assertTrue(callable(self.analysisstep.add_preprocessing_step))

    def test_add_processing_step_copies_processingstep_object(self):
        self.analysisstep.add_preprocessing_step(self.processingstep)
        self.assertIsNot(
            self.processingstep, self.analysisstep.preprocessing[-1]
        )


class TestBasicCharacteristics(unittest.TestCase):
    def setUp(self):
        self.analysis = aspecd.analysis.BasicCharacteristics()
        self.dataset = aspecd.dataset.Dataset()
        self.dataset.data.data = np.random.random(10)
        self.dataset3d = aspecd.dataset.Dataset()
        self.dataset3d.data.data = np.random.random([10, 10, 10])

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn(
            "basic characteristics", self.analysis.description.lower()
        )

    def test_analyse_without_kind_raises(self):
        with self.assertRaisesRegex(
            ValueError, "No kind of characteristics given"
        ):
            self.dataset.analyse(self.analysis)

    def test_analyse_with_unknown_kind_raises(self):
        self.analysis.parameters["kind"] = "foo"
        with self.assertRaisesRegex(ValueError, "Unknown kind foo"):
            self.dataset.analyse(self.analysis)

    def test_analyse_with_unknown_output_raises(self):
        self.analysis.parameters["kind"] = "min"
        self.analysis.parameters["output"] = "foo"
        with self.assertRaisesRegex(ValueError, "Unknown output type foo"):
            self.dataset.analyse(self.analysis)

    def test_min(self):
        self.analysis.parameters["kind"] = "min"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(self.dataset.data.data.min(), analysis.result)

    def test_min_sets_kind_in_index(self):
        self.analysis.parameters["kind"] = "min"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual([self.analysis.parameters["kind"]], analysis.index)

    def test_max(self):
        self.analysis.parameters["kind"] = "max"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(self.dataset.data.data.max(), analysis.result)

    def test_amplitude(self):
        self.analysis.parameters["kind"] = "amplitude"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(
            self.dataset.data.data.max() - self.dataset.data.data.min(),
            analysis.result,
        )

    def test_area(self):
        self.analysis.parameters["kind"] = "area"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(self.dataset.data.data.sum(), analysis.result)

    def test_all(self):
        self.analysis.parameters["kind"] = "all"
        result_dict = {
            "min": self.dataset.data.data.min(),
            "max": self.dataset.data.data.max(),
            "amplitude": self.dataset.data.data.max()
            - self.dataset.data.data.min(),
            "area": self.dataset.data.data.sum(),
        }
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(list(result_dict.values()), analysis.result)
        self.assertEqual(list(result_dict.keys()), analysis.index)

    def test_min_with_axes_output(self):
        self.analysis.parameters["kind"] = "min"
        self.analysis.parameters["output"] = "axes"
        analysis = self.dataset.analyse(self.analysis)
        result = [
            self.dataset.data.axes[0].values[self.dataset.data.data.argmin()]
        ]
        self.assertEqual(result, analysis.result)

    def test_min_with_axes_output_sets_index(self):
        self.analysis.parameters["kind"] = "min"
        self.analysis.parameters["output"] = "axes"
        self.dataset.data.axes[0].quantity = "foo"
        analysis = self.dataset.analyse(self.analysis)
        index = ["min(foo)"]
        self.assertEqual(index, analysis.index)

    def test_max_with_axes_output(self):
        self.analysis.parameters["kind"] = "max"
        self.analysis.parameters["output"] = "axes"
        analysis = self.dataset.analyse(self.analysis)
        result = [
            self.dataset.data.axes[0].values[self.dataset.data.data.argmax()]
        ]
        self.assertEqual(result, analysis.result)

    def test_max_with_axes_output_sets_index(self):
        self.analysis.parameters["kind"] = "max"
        self.analysis.parameters["output"] = "axes"
        self.dataset.data.axes[0].quantity = "foo"
        analysis = self.dataset.analyse(self.analysis)
        index = ["max(foo)"]
        self.assertEqual(index, analysis.index)

    def test_min_with_indices_output(self):
        self.analysis.parameters["kind"] = "min"
        self.analysis.parameters["output"] = "indices"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual([self.dataset.data.data.argmin()], analysis.result)

    def test_min_with_indices_output_sets_index(self):
        self.analysis.parameters["kind"] = "min"
        self.analysis.parameters["output"] = "indices"
        analysis = self.dataset.analyse(self.analysis)
        index = ["min(index0)"]
        self.assertEqual(index, analysis.index)

    def test_max_with_indices_output(self):
        self.analysis.parameters["kind"] = "max"
        self.analysis.parameters["output"] = "indices"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual([self.dataset.data.data.argmax()], analysis.result)

    def test_max_with_indices_output_sets_index(self):
        self.analysis.parameters["kind"] = "max"
        self.analysis.parameters["output"] = "indices"
        analysis = self.dataset.analyse(self.analysis)
        index = ["max(index0)"]
        self.assertEqual(index, analysis.index)

    def test_area_or_amplitude_with_axes_output_raises(self):
        characteristics = ["area", "amplitude"]
        self.analysis.parameters["output"] = "axes"
        for characteristic in characteristics:
            with self.subTest(characteristic=characteristic):
                self.analysis.parameters["kind"] = characteristic
                with self.assertRaisesRegex(
                    ValueError,
                    "Output %s not available for "
                    "characteristic %s." % ("axes", characteristic),
                ):
                    self.dataset.analyse(self.analysis)

    def test_area_or_amplitude_with_indices_output_raises(self):
        characteristics = ["area", "amplitude"]
        self.analysis.parameters["output"] = "indices"
        for characteristic in characteristics:
            with self.subTest(characteristic=characteristic):
                self.analysis.parameters["kind"] = characteristic
                with self.assertRaisesRegex(
                    ValueError,
                    "Output %s not available for "
                    "characteristic %s." % ("indices", characteristic),
                ):
                    self.dataset.analyse(self.analysis)

    def test_min_with_3d_data(self):
        self.analysis.parameters["kind"] = "min"
        analysis = self.dataset3d.analyse(self.analysis)
        self.assertEqual(self.dataset3d.data.data.min(), analysis.result)

    def test_max_with_3d_data(self):
        self.analysis.parameters["kind"] = "max"
        analysis = self.dataset3d.analyse(self.analysis)
        self.assertEqual(self.dataset3d.data.data.max(), analysis.result)

    def test_amplitude_with_3d_data(self):
        self.analysis.parameters["kind"] = "amplitude"
        analysis = self.dataset3d.analyse(self.analysis)
        self.assertEqual(
            self.dataset3d.data.data.max() - self.dataset3d.data.data.min(),
            analysis.result,
        )

    def test_area_with_3d_data(self):
        self.analysis.parameters["kind"] = "area"
        analysis = self.dataset3d.analyse(self.analysis)
        self.assertEqual(self.dataset3d.data.data.sum(), analysis.result)

    def test_all_with_3d_data(self):
        self.analysis.parameters["kind"] = "all"
        result_dict = {
            "min": self.dataset3d.data.data.min(),
            "max": self.dataset3d.data.data.max(),
            "amplitude": self.dataset3d.data.data.max()
            - self.dataset3d.data.data.min(),
            "area": self.dataset3d.data.data.sum(),
        }
        analysis = self.dataset3d.analyse(self.analysis)
        self.assertEqual(list(result_dict.values()), analysis.result)

    def test_min_with_axes_output_with_3d_data(self):
        self.analysis.parameters["kind"] = "min"
        self.analysis.parameters["output"] = "axes"
        analysis = self.dataset3d.analyse(self.analysis)
        result = []
        idx = np.unravel_index(
            self.dataset3d.data.data.argmin(), self.dataset3d.data.data.shape
        )
        for dim in range(self.dataset3d.data.data.ndim):
            result.append(self.dataset3d.data.axes[dim].values[idx[dim]])
        self.assertEqual(result, analysis.result)

    def test_max_with_axes_output_with_3d_data(self):
        self.analysis.parameters["kind"] = "max"
        self.analysis.parameters["output"] = "axes"
        analysis = self.dataset3d.analyse(self.analysis)
        result = []
        idx = np.unravel_index(
            self.dataset3d.data.data.argmax(), self.dataset3d.data.data.shape
        )
        for dim in range(self.dataset3d.data.data.ndim):
            result.append(self.dataset3d.data.axes[dim].values[idx[dim]])
        self.assertEqual(result, analysis.result)

    def test_min_with_indices_output_with_3d_data(self):
        self.analysis.parameters["kind"] = "min"
        self.analysis.parameters["output"] = "indices"
        analysis = self.dataset3d.analyse(self.analysis)
        result = list(
            np.unravel_index(
                self.dataset3d.data.data.argmin(),
                self.dataset3d.data.data.shape,
            )
        )
        self.assertListEqual(result, analysis.result)

    def test_max_with_indices_output_with_3d_data(self):
        self.analysis.parameters["kind"] = "max"
        self.analysis.parameters["output"] = "indices"
        analysis = self.dataset3d.analyse(self.analysis)
        result = list(
            np.unravel_index(
                self.dataset3d.data.data.argmax(),
                self.dataset3d.data.data.shape,
            )
        )
        self.assertListEqual(result, analysis.result)


class TestBasicStatistics(unittest.TestCase):
    def setUp(self):
        self.analysis = aspecd.analysis.BasicStatistics()
        self.dataset = aspecd.dataset.Dataset()
        self.dataset.data.data = np.random.random(10)
        self.dataset3d = aspecd.dataset.Dataset()
        self.dataset3d.data.data = np.random.random([10, 10, 10])

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("basic statistics", self.analysis.description.lower())

    def test_analyse_without_kind_raises(self):
        with self.assertRaisesRegex(
            ValueError, "No kind of statistics given"
        ):
            self.dataset.analyse(self.analysis)

    def test_analyse_with_unknown_kind_raises(self):
        self.analysis.parameters["kind"] = "foo"
        with self.assertRaisesRegex(ValueError, "Unknown kind foo"):
            self.dataset.analyse(self.analysis)

    def test_mean_with_1d_data(self):
        self.analysis.parameters["kind"] = "mean"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(np.mean(self.dataset.data.data), analysis.result)

    def test_median_with_1d_data(self):
        self.analysis.parameters["kind"] = "median"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(np.median(self.dataset.data.data), analysis.result)

    def test_std_with_1d_data(self):
        self.analysis.parameters["kind"] = "std"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(np.std(self.dataset.data.data), analysis.result)

    def test_var_with_1d_data(self):
        self.analysis.parameters["kind"] = "var"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(np.var(self.dataset.data.data), analysis.result)

    def test_mean_with_3d_data(self):
        self.analysis.parameters["kind"] = "mean"
        analysis = self.dataset3d.analyse(self.analysis)
        self.assertEqual(np.mean(self.dataset3d.data.data), analysis.result)

    def test_median_with_3d_data(self):
        self.analysis.parameters["kind"] = "median"
        analysis = self.dataset3d.analyse(self.analysis)
        self.assertEqual(np.median(self.dataset3d.data.data), analysis.result)

    def test_std_with_3d_data(self):
        self.analysis.parameters["kind"] = "std"
        analysis = self.dataset3d.analyse(self.analysis)
        self.assertEqual(np.std(self.dataset3d.data.data), analysis.result)

    def test_var_with_3d_data(self):
        self.analysis.parameters["kind"] = "var"
        analysis = self.dataset3d.analyse(self.analysis)
        self.assertEqual(np.var(self.dataset3d.data.data), analysis.result)


class TestBlindSNREstimation(unittest.TestCase):
    def setUp(self):
        self.analysis = aspecd.analysis.BlindSNREstimation()
        self.dataset = aspecd.dataset.Dataset()
        self.dataset.data.data = np.random.random(10)

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn(
            "blind signal-to-noise ratio estimation",
            self.analysis.description.lower(),
        )

    def test_analyse_without_method_sets_method_to_simple(self):
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(analysis.parameters["method"], "simple")

    def test_simple_method(self):
        analysis = self.dataset.analyse(self.analysis)
        result = self.dataset.data.data.mean() / self.dataset.data.data.std()
        self.assertEqual(result, analysis.result)

    def test_simple_squared_method(self):
        self.analysis.parameters["method"] = "simple_squared"
        analysis = self.dataset.analyse(self.analysis)
        result = (
            self.dataset.data.data.mean() ** 2
            / self.dataset.data.data.std() ** 2
        )
        self.assertEqual(result, analysis.result)

    def test_der_snr_method_with_too_small_sample_raises(self):
        self.dataset.data.data = np.random.random(3)
        self.analysis.parameters["method"] = "der_snr"
        with self.assertRaisesRegex(ValueError, "Too few samples"):
            self.dataset.analyse(self.analysis)

    def test_der_snr_method(self):
        self.analysis.parameters["method"] = "der_snr"
        analysis = self.dataset.analyse(self.analysis)
        self.assertGreater(analysis.result, 0)


class TestPeakFinding(unittest.TestCase):
    def setUp(self):
        self.analysis = aspecd.analysis.PeakFinding()
        self.dataset = aspecd.dataset.Dataset()
        self.dataset.data.data = np.sin(np.linspace(0, 8 * np.pi, num=1000))
        self.noisy_dataset = aspecd.dataset.Dataset()
        self.noisy_dataset.data.data = np.sin(
            np.linspace(0, 8 * np.pi, num=1000)
        )
        self.noisy_dataset.data.data += (
            np.random.random(len(self.noisy_dataset.data.data)) - 0.5
        ) * 0.2

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("peak finding", self.analysis.description.lower())

    def test_with_nd_dataset_raises(self):
        self.dataset.data.data = np.random.random([5, 5])
        with self.assertRaises(aspecd.exceptions.NotApplicableToDatasetError):
            self.dataset.analyse(self.analysis)

    def test_analyse_returns_peak_positions(self):
        analysis = self.dataset.analyse(self.analysis)
        result, _ = scipy.signal.find_peaks(self.dataset.data.data)
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_returns_peak_positions_in_axis_values(self):
        self.dataset.data.axes[0].values = np.linspace(
            340, 350, len(self.dataset.data.data)
        )
        analysis = self.dataset.analyse(self.analysis)
        result, _ = scipy.signal.find_peaks(self.dataset.data.data)
        result = self.dataset.data.axes[0].values[result]
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_returns_properties(self):
        self.analysis.parameters["return_properties"] = True
        analysis = self.dataset.analyse(self.analysis)
        result, _ = scipy.signal.find_peaks(self.dataset.data.data)
        self.assertListEqual(list(result), list(analysis.result[0]))
        self.assertEqual(dict, type(analysis.result[1]))

    def test_analyse_returns_dataset(self):
        self.analysis.parameters["return_dataset"] = True
        analysis = self.dataset.analyse(self.analysis)
        result, _ = scipy.signal.find_peaks(self.dataset.data.data)
        self.assertListEqual(
            list(result), list(analysis.result.data.axes[0].values)
        )
        self.assertEqual(
            aspecd.dataset.CalculatedDataset, type(analysis.result)
        )
        self.assertEqual(
            self.analysis.name, analysis.result.metadata.calculation.type
        )
        self.assertDictEqual(
            self.analysis.parameters,
            analysis.result.metadata.calculation.parameters,
        )

    def test_analyse_returns_negative_peak_positions(self):
        self.analysis.parameters["negative_peaks"] = True
        analysis = self.dataset.analyse(self.analysis)
        positive, _ = scipy.signal.find_peaks(self.dataset.data.data)
        negative, _ = scipy.signal.find_peaks(-self.dataset.data.data)
        result = np.sort(np.concatenate((positive, negative)))
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_negative_peaks_does_not_return_properties(self):
        self.analysis.parameters["negative_peaks"] = True
        self.analysis.parameters["return_properties"] = True
        analysis = self.dataset.analyse(self.analysis)
        positive, _ = scipy.signal.find_peaks(self.dataset.data.data)
        negative, _ = scipy.signal.find_peaks(-self.dataset.data.data)
        result = np.sort(np.concatenate((positive, negative)))
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_height(self):
        self.analysis.parameters["height"] = 1
        analysis = self.noisy_dataset.analyse(self.analysis)
        result, _ = scipy.signal.find_peaks(
            self.noisy_dataset.data.data,
            height=self.analysis.parameters["height"],
        )
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_height_with_negative_peak_positions(self):
        self.analysis.parameters["negative_peaks"] = True
        self.analysis.parameters["height"] = 1
        analysis = self.noisy_dataset.analyse(self.analysis)
        positive, _ = scipy.signal.find_peaks(
            self.noisy_dataset.data.data,
            height=self.analysis.parameters["height"],
        )
        negative, _ = scipy.signal.find_peaks(
            -self.noisy_dataset.data.data,
            height=self.analysis.parameters["height"],
        )
        result = np.sort(np.concatenate((positive, negative)))
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_threshold(self):
        self.analysis.parameters["threshold"] = 0.145
        analysis = self.noisy_dataset.analyse(self.analysis)
        result, _ = scipy.signal.find_peaks(
            self.noisy_dataset.data.data,
            threshold=self.analysis.parameters["threshold"],
        )
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_threshold_with_negative_peak_positions(self):
        self.analysis.parameters["negative_peaks"] = True
        self.analysis.parameters["threshold"] = 0.145
        analysis = self.noisy_dataset.analyse(self.analysis)
        positive, _ = scipy.signal.find_peaks(
            self.noisy_dataset.data.data,
            threshold=self.analysis.parameters["threshold"],
        )
        negative, _ = scipy.signal.find_peaks(
            -self.noisy_dataset.data.data,
            threshold=self.analysis.parameters["threshold"],
        )
        result = np.sort(np.concatenate((positive, negative)))
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_distance(self):
        self.analysis.parameters["distance"] = 100
        analysis = self.noisy_dataset.analyse(self.analysis)
        result, _ = scipy.signal.find_peaks(
            self.noisy_dataset.data.data,
            distance=self.analysis.parameters["distance"],
        )
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_distance_with_negative_peak_positions(self):
        self.analysis.parameters["negative_peaks"] = True
        self.analysis.parameters["distance"] = 100
        analysis = self.noisy_dataset.analyse(self.analysis)
        positive, _ = scipy.signal.find_peaks(
            self.noisy_dataset.data.data,
            distance=self.analysis.parameters["distance"],
        )
        negative, _ = scipy.signal.find_peaks(
            -self.noisy_dataset.data.data,
            distance=self.analysis.parameters["distance"],
        )
        result = np.sort(np.concatenate((positive, negative)))
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_prominence(self):
        self.analysis.parameters["prominence"] = 0.2
        analysis = self.noisy_dataset.analyse(self.analysis)
        result, _ = scipy.signal.find_peaks(
            self.noisy_dataset.data.data,
            prominence=self.analysis.parameters["prominence"],
        )
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_prominence_with_negative_peak_positions(self):
        self.analysis.parameters["negative_peaks"] = True
        self.analysis.parameters["prominence"] = 0.2
        analysis = self.noisy_dataset.analyse(self.analysis)
        positive, _ = scipy.signal.find_peaks(
            self.noisy_dataset.data.data,
            prominence=self.analysis.parameters["prominence"],
        )
        negative, _ = scipy.signal.find_peaks(
            -self.noisy_dataset.data.data,
            prominence=self.analysis.parameters["prominence"],
        )
        result = np.sort(np.concatenate((positive, negative)))
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_width(self):
        self.analysis.parameters["width"] = 5
        analysis = self.noisy_dataset.analyse(self.analysis)
        result, _ = scipy.signal.find_peaks(
            self.noisy_dataset.data.data,
            width=self.analysis.parameters["width"],
        )
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_with_width_with_negative_peak_positions(self):
        self.analysis.parameters["negative_peaks"] = True
        self.analysis.parameters["width"] = 5
        analysis = self.noisy_dataset.analyse(self.analysis)
        positive, _ = scipy.signal.find_peaks(
            self.noisy_dataset.data.data,
            width=self.analysis.parameters["width"],
        )
        negative, _ = scipy.signal.find_peaks(
            -self.noisy_dataset.data.data,
            width=self.analysis.parameters["width"],
        )
        result = np.sort(np.concatenate((positive, negative)))
        self.assertListEqual(list(result), list(analysis.result))

    def test_analyse_returns_peak_positions_and_intensities(self):
        self.analysis.parameters["return_intensities"] = True
        self.dataset.data.axes[0].values = np.linspace(
            340, 350, len(self.dataset.data.data)
        )
        analysis = self.dataset.analyse(self.analysis)
        result, _ = scipy.signal.find_peaks(self.dataset.data.data)
        heights = self.dataset.data.data[result]
        result = self.dataset.data.axes[0].values[result]
        np.testing.assert_array_equal(
            np.stack((result, heights), axis=1), analysis.result
        )


class TestPowerDensitySpectrum(unittest.TestCase):
    def setUp(self):
        self.analysis = aspecd.analysis.PowerDensitySpectrum()
        self.dataset = aspecd.dataset.Dataset()
        self.dataset.data.data = np.zeros(2**12)
        noise = aspecd.processing.Noise()
        self.dataset.process(noise)

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn(
            "power density spectrum", self.analysis.description.lower()
        )

    def test_with_nd_dataset_raises(self):
        self.dataset.data.data = np.random.random([5, 5])
        with self.assertRaises(aspecd.exceptions.NotApplicableToDatasetError):
            self.dataset.analyse(self.analysis)

    def test_returns_calculated_dataset(self):
        analysis = self.dataset.analyse(self.analysis)
        self.assertIsInstance(
            analysis.result, aspecd.dataset.CalculatedDataset
        )

    def test_returned_data_are_log_of_psd(self):
        frequencies, psd = scipy.signal.periodogram(self.dataset.data.data)
        log_psd = np.log10(psd[1:])
        analysis = self.dataset.analyse(self.analysis)
        self.assertListEqual(list(log_psd), list(analysis.result.data.data))

    def test_returned_axes_values_are_log_of_frequency(self):
        frequencies, psd = scipy.signal.periodogram(self.dataset.data.data)
        log_frequencies = np.log10(frequencies[1:])
        analysis = self.dataset.analyse(self.analysis)
        self.assertListEqual(
            list(log_frequencies), list(analysis.result.data.axes[0].values)
        )

    def test_axes_quantities_are_sensible(self):
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(
            "log frequency", analysis.result.data.axes[0].quantity
        )
        self.assertEqual("log power", analysis.result.data.axes[1].quantity)

    def test_method_settable_via_parameter(self):
        self.analysis.parameters["method"] = "welch"
        frequencies, psd = scipy.signal.welch(self.dataset.data.data)
        log_psd = np.log10(psd[1:])
        analysis = self.dataset.analyse(self.analysis)
        self.assertListEqual(list(log_psd), list(analysis.result.data.data))

    def test_method_defaults_to_periodogram(self):
        self.assertEqual("periodogram", self.analysis.parameters["method"])


class TestPolynomialFit(unittest.TestCase):
    def setUp(self):
        self.analysis = aspecd.analysis.PolynomialFit()
        self.dataset = aspecd.dataset.Dataset()
        self.dataset.data.data = np.linspace(0, 49)

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("polynomial fit", self.analysis.description.lower())

    def test_with_nd_dataset_raises(self):
        self.dataset.data.data = np.random.random([5, 5])
        with self.assertRaises(aspecd.exceptions.NotApplicableToDatasetError):
            self.dataset.analyse(self.analysis)

    def test_default_is_first_order(self):
        self.assertEqual(1, self.analysis.parameters["order"])

    def test_first_order_returns_two_coefficients_as_result(self):
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(2, len(analysis.result))

    def test_first_order_returns_correct_coefficients(self):
        analysis = self.dataset.analyse(self.analysis)
        self.assertAlmostEqual(0, analysis.result[0])
        self.assertAlmostEqual(1, analysis.result[1])

    def test_zeroth_order_returns_correct_coefficients(self):
        self.analysis.parameters["order"] = 0
        analysis = self.dataset.analyse(self.analysis)
        self.assertAlmostEqual(24.5, analysis.result[0])


class TestLinearRegressionWithFixedIntercept(unittest.TestCase):
    def setUp(self):
        self.analysis = aspecd.analysis.LinearRegressionWithFixedIntercept()
        self.dataset = aspecd.dataset.Dataset()
        self.dataset.data.data = np.linspace(1, 50)
        self.dataset.data.axes[0].values = np.linspace(0.5, 25)

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn(
            "linear regression without intercept",
            self.analysis.description.lower(),
        )

    def test_with_nd_dataset_raises(self):
        self.dataset.data.data = np.random.random([5, 5])
        with self.assertRaises(aspecd.exceptions.NotApplicableToDatasetError):
            self.dataset.analyse(self.analysis)

    def test_analysis_returns_correct_coefficient_as_result(self):
        analysis = self.dataset.analyse(self.analysis)
        self.assertAlmostEqual(2.0, analysis.result)

    def test_regression_with_constant_offset(self):
        self.dataset.data.data += np.pi
        self.analysis.parameters["offset"] = np.pi
        analysis = self.dataset.analyse(self.analysis)
        self.assertAlmostEqual(2.0, analysis.result)

    def test_polynomial_compatible_coefficients(self):
        self.dataset.data.data += np.pi
        self.analysis.parameters["offset"] = np.pi
        self.analysis.parameters["polynomial_coefficients"] = True
        analysis = self.dataset.analyse(self.analysis)
        self.assertAlmostEqual(np.pi, analysis.result[0])
        self.assertAlmostEqual(2.0, analysis.result[1])


class TestDeviceDataExtraction(unittest.TestCase):
    def setUp(self):
        self.analysis = aspecd.analysis.DeviceDataExtraction()
        self.dataset = aspecd.dataset.Dataset()
        device_data = aspecd.dataset.DeviceData()
        device_data.data = np.linspace(1, 50)
        device_data.axes[0].values = np.linspace(0.5, 25)
        self.dataset.device_data = {"test": device_data}
        self.analysis.parameters["device"] = "test"

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn(
            "extract device data", self.analysis.description.lower()
        )

    def test_analysis_returns_dataset(self):
        analysis = self.dataset.analyse(self.analysis)
        self.assertTrue(isinstance(analysis.result, aspecd.dataset.Dataset))

    def test_analysis_with_unknown_device_raises(self):
        self.analysis.parameters["device"] = "nonexisting_device"
        with self.assertRaises(KeyError):
            self.dataset.analyse(self.analysis)

    def test_analysis_returns_dataset_with_correct_data(self):
        analysis = self.dataset.analyse(self.analysis)
        self.assertListEqual(
            list(analysis.result.data.data),
            list(self.dataset.device_data["test"].data),
        )

    def test_returned_dataset_axes_are_not_identical_to_device_axes(self):
        analysis = self.dataset.analyse(self.analysis)
        # Hint: comparing data rather than axes does not work
        self.assertIsNot(
            analysis.result.data.axes, self.dataset.device_data["test"].axes
        )

    def test_returned_dataset_contains_reference_to_original_dataset(self):
        self.dataset.id = "My unique ID"
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(analysis.result.references[0].id, self.dataset.id)

    def test_analysis_with_dataset_with_no_device_data_raises(self):
        self.dataset.device_data = {}
        with self.assertRaises(aspecd.exceptions.NotApplicableToDatasetError):
            self.dataset.analyse(self.analysis)


class TestCentreOfMass(unittest.TestCase):
    def setUp(self):
        self.analysis = aspecd.analysis.CentreOfMass()
        self.dataset = aspecd.dataset.Dataset()
        self.dataset.data.data = np.sin(np.linspace(0, np.pi, num=1001))

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn("centre of mass", self.analysis.description.lower())

    def test_analyse_returns_centre_of_mass_in_1d_with_indices(self):
        analysis = self.dataset.analyse(self.analysis)
        np.testing.assert_almost_equal(500, analysis.result)

    def test_analyse_returns_centre_of_mass_in_1d_with_axis_values(self):
        self.dataset.data.axes[0].values = np.linspace(
            340, 350, len(self.dataset.data.data)
        )
        analysis = self.dataset.analyse(self.analysis)
        np.testing.assert_almost_equal(345, analysis.result)

    def test_analyse_returns_centre_of_mass_in_2d_with_indices(self):
        self.dataset.data.data = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 1, 1, 0],
                [0, 1, 2, 1, 0],
                [0, 1, 1, 1, 0],
                [0, 0, 0, 0, 0],
            ]
        )
        analysis = self.dataset.analyse(self.analysis)
        np.testing.assert_array_almost_equal(
            np.array([2, 2]), analysis.result
        )

    def test_analyse_returns_centre_of_mass_in_2d_with_axis_values(self):
        self.dataset.data.data = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 1, 1, 0],
                [0, 1, 2, 1, 0],
                [0, 1, 1, 1, 0],
                [0, 0, 0, 0, 0],
            ]
        )
        self.dataset.data.axes[0].values = np.linspace(340, 350, 5)
        self.dataset.data.axes[1].values = np.linspace(230, 235, 5)
        analysis = self.dataset.analyse(self.analysis)
        np.testing.assert_array_almost_equal(
            np.array([345, 232.5]), analysis.result
        )



import os
import setuptools


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        content = f.read()
    return content


setuptools.setup(
    name="aspecd",
    version=read("VERSION").strip(),
    description="Framework for handling spectroscopic data.",
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    author="Till Biskup",
    author_email="till@till-biskup.de",
    url="https://www.aspecd.de/",
    project_urls={
        "Documentation": "https://docs.aspecd.de/",
        "Source": "https://github.com/tillbiskup/aspecd",
    },
    packages=setuptools.find_packages(exclude=("tests", "docs")),
    license="BSD",
    keywords=[
        "spectroscopy",
        "data processing and analysis",
        "reproducible science",
        "reproducible research",
        "good scientific practice",
        "recipe-driven data analysis",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
    install_requires=[
        "jinja2>=3.0",
        "matplotlib",
        "numpy",
        "scipy>=0.14",
        "oyaml",
        "asdf",
        "bibrecord",
    ],
    extras_require={
        "dev": [
            "prospector",
            "pyroma",
            "bandit",
            "black",
        ],
        "docs": [
            "sphinx",
            "sphinx-rtd-theme",
            "sphinx-multiversion",
        ],
        "deployment": [
            "build",
            "twine",
        ],
    },
    python_requires=">=3.7",
    entry_points={
        "console_scripts": [
            "serve = aspecd.tasks:serve",
        ],
    },
    include_package_data=True,
)
